// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "atlres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// Chinese (Simplified, PRC) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_CHS)
LANGUAGE LANG_CHINESE, SUBLANG_CHINESE_SIMPLIFIED
#pragma code_page(936)

/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_DLG_SERIAL_SEL DIALOGEX 0, 0, 244, 72
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Select Serial Port..."
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,127,50,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,187,50,50,14
    COMBOBOX        IDC_COMB_SERIAL_SEL,7,7,230,88,CBS_DROPDOWNLIST | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "TCP Server",IDC_BUTTON_TCPSERVER,7,28,50,14
    COMBOBOX        IDC_COMB_BAUDRATE,162,28,75,30,CBS_DROPDOWN | WS_VSCROLL | WS_TABSTOP
END

IDD_DLG_SETFREQ DIALOGEX 0, 0, 332, 100
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Frequence"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "Set",IDOK,171,15,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,275,79,50,14
    CONTROL         "",IDC_SLIDER_FREQ,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,83,36,152,21
    LTEXT           "Min(0)",IDC_STATIC_MINFREQ,63,40,20,8
    LTEXT           "Max(1023)",IDC_STATIC_MAXFREQ,236,40,40,8
    EDITTEXT        IDC_EDIT_FREQ,119,15,40,14,ES_AUTOHSCROLL | ES_NUMBER
END

IDD_DIG_TCP_SEL DIALOGEX 0, 0, 252, 78
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "TCP Address"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,132,49,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,186,49,50,14
    LTEXT           "IP:",IDC_STATIC,21,20,10,8
    EDITTEXT        IDC_EDIT_IP,44,19,72,14,ES_AUTOHSCROLL
    LTEXT           "Port:",IDC_STATIC,139,20,17,8
    EDITTEXT        IDC_EDIT_PORT,166,19,70,14,ES_AUTOHSCROLL
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    IDD_DLG_SERIAL_SEL, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 237
        TOPMARGIN, 7
        BOTTOMMARGIN, 65
        HORZGUIDE, 28
    END

    IDD_DLG_SETFREQ, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 325
        TOPMARGIN, 7
        BOTTOMMARGIN, 93
    END

    IDD_DIG_TCP_SEL, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 245
        TOPMARGIN, 7
        BOTTOMMARGIN, 71
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// AFX_DIALOG_LAYOUT
//

IDD_DLG_SERIAL_SEL AFX_DIALOG_LAYOUT
BEGIN
    0
END

#endif    // Chinese (Simplified, PRC) resources
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
// English (United States) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)

/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_ABOUTBOX DIALOGEX 0, 0, 233, 94
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "About"
FONT 8, "MS Shell Dlg", 0, 0, 0x0
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,176,73,50,14
    LTEXT           "RPLIDAR(R) System Win32 Demo Application\n\nVersion: 1.10.0\n\nCopyright(C) 2009 - 2013 RoboPeak Team\nCopyright(C) 2013 - 2019 Shanghai Slamtec Co. Ltd.\nhttp://www.slamtec.com",IDC_STATIC,45,14,172,57
    ICON            IDR_MAINFRAME,IDC_STATIC,17,15,20,20
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    IDD_ABOUTBOX, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 226
        TOPMARGIN, 7
        BOTTOMMARGIN, 87
    END
END
#endif    // APSTUDIO_INVOKED


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#include ""atlres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE 
BEGIN
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Toolbar
//

IDR_MAINFRAME TOOLBAR 16, 15
BEGIN
    BUTTON      ID_FILE_DUMPDATA
    SEPARATOR
    BUTTON      ID_CMD_STOP
    SEPARATOR
    BUTTON      ID_CMD_SCAN
    SEPARATOR
    BUTTON      ID_CMD_RESET
    SEPARATOR
    BUTTON      ID_CMD_SET_FREQ
    BUTTON      ID_APP_ABOUT
END


/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDR_MAINFRAME           BITMAP                  "res\\Toolbar.bmp"


/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDR_MAINFRAME MENU
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "Dump Data...",                ID_FILE_DUMPDATA
        MENUITEM SEPARATOR
        MENUITEM "E&xit",                       ID_APP_EXIT
    END
    POPUP "&Command"
    BEGIN
        MENUITEM "STOP",                        ID_CMD_STOP
        MENUITEM "Scan",                        ID_CMD_SCAN
        MENUITEM SEPARATOR
        MENUITEM "Reset",                       ID_CMD_RESET
    END
    POPUP "&Option"
    BEGIN
        MENUITEM "Force Scan",                  ID_OPT_FORCESCAN
    END
    POPUP "&View"
    BEGIN
        MENUITEM "&Toolbar",                    ID_VIEW_TOOLBAR
        MENUITEM "&Status Bar",                 ID_VIEW_STATUS_BAR
    END
    POPUP "&Help"
    BEGIN
        MENUITEM "&About RoboPeak Lidar Tool",  ID_APP_ABOUT
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDR_MAINFRAME           ICON                    "res\\rplidar.ico"


/////////////////////////////////////////////////////////////////////////////
//
// Accelerator
//

IDR_MAINFRAME ACCELERATORS
BEGIN
    "N",            ID_FILE_NEW,            VIRTKEY, CONTROL
    "O",            ID_FILE_OPEN,           VIRTKEY, CONTROL
    "S",            ID_FILE_SAVE,           VIRTKEY, CONTROL
    "P",            ID_FILE_PRINT,          VIRTKEY, CONTROL
    "Z",            ID_EDIT_UNDO,           VIRTKEY, CONTROL
    "X",            ID_EDIT_CUT,            VIRTKEY, CONTROL
    "C",            ID_EDIT_COPY,           VIRTKEY, CONTROL
    "V",            ID_EDIT_PASTE,          VIRTKEY, CONTROL
    VK_BACK,        ID_EDIT_UNDO,           VIRTKEY, ALT
    VK_DELETE,      ID_EDIT_CUT,            VIRTKEY, SHIFT
    VK_INSERT,      ID_EDIT_COPY,           VIRTKEY, CONTROL
    VK_INSERT,      ID_EDIT_PASTE,          VIRTKEY, SHIFT
    VK_F6,          ID_NEXT_PANE,           VIRTKEY 
    VK_F6,          ID_PREV_PANE,           VIRTKEY, SHIFT
END


/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 1,10,0,0
 PRODUCTVERSION 1,10,0,0
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x4L
 FILETYPE 0x2L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "FileDescription", "framegrabber Module"
            VALUE "FileVersion", "1.10.0.0"
            VALUE "InternalName", "framegrabber"
            VALUE "LegalCopyright", "Copyright 2012"
            VALUE "OriginalFilename", "framegrabber.exe"
            VALUE "ProductName", "framegrabber Module"
            VALUE "ProductVersion", "1.10.0.0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// AFX_DIALOG_LAYOUT
//

IDD_ABOUTBOX AFX_DIALOG_LAYOUT
BEGIN
    0
END


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE
BEGIN
    IDR_MAINFRAME           "framegrabber"
END

STRINGTABLE
BEGIN
    ID_FILE_NEW             "Create a new document\nNew"
    ID_FILE_OPEN            "Open an existing document\nOpen"
    ID_FILE_CLOSE           "Close the active document\nClose"
    ID_FILE_SAVE            "Save the active document\nSave"
    ID_FILE_SAVE_AS         "Save the active document with a new name\nSave As"
    ID_FILE_PAGE_SETUP      "Change the printing options\nPage Setup"
    ID_FILE_PRINT_SETUP     "Change the printer and printing options\nPrint Setup"
    ID_FILE_PRINT           "Print the active document\nPrint"
    ID_FILE_PRINT_PREVIEW   "Display full pages\nPrint Preview"
END

STRINGTABLE
BEGIN
    ID_APP_ABOUT            "About\nAbout"
    ID_APP_EXIT             "Quit the application; prompts to save documents\nExit"
END

STRINGTABLE
BEGIN
    ID_NEXT_PANE            "Switch to the next window pane\nNext Pane"
    ID_PREV_PANE            "Switch back to the previous window pane\nPrevious Pane"
END

STRINGTABLE
BEGIN
    ID_WINDOW_NEW           "Open another window for the active document\nNew Window"
    ID_WINDOW_ARRANGE       "Arrange icons at the bottom of the window\nArrange Icons"
    ID_WINDOW_CASCADE       "Arrange windows so they overlap\nCascade Windows"
    ID_WINDOW_TILE_HORZ     "Arrange windows as non-overlapping tiles\nTile Windows"
    ID_WINDOW_TILE_VERT     "Arrange windows as non-overlapping tiles\nTile Windows"
    ID_WINDOW_SPLIT         "Split the active window into panes\nSplit"
END

STRINGTABLE
BEGIN
    ID_EDIT_CLEAR           "Erase the selection\nErase"
    ID_EDIT_CLEAR_ALL       "Erase everything\nErase All"
    ID_EDIT_COPY            "Copy the selection and put it on the Clipboard\nCopy"
    ID_EDIT_CUT             "Cut the selection and put it on the Clipboard\nCut"
    ID_EDIT_FIND            "Find the specified text\nFind"
    ID_EDIT_PASTE           "Insert Clipboard contents\nPaste"
    ID_EDIT_REPEAT          "Repeat the last action\nRepeat"
    ID_EDIT_REPLACE         "Replace specific text with different text\nReplace"
    ID_EDIT_SELECT_ALL      "Select the entire document\nSelect All"
    ID_EDIT_UNDO            "Undo the last action\nUndo"
    ID_EDIT_REDO            "Redo the previously undone action\nRedo"
END

STRINGTABLE
BEGIN
    ID_VIEW_TOOLBAR         "Show or hide the toolbar\nToggle ToolBar"
    ID_VIEW_STATUS_BAR      "Show or hide the status bar\nToggle StatusBar"
END

STRINGTABLE
BEGIN
    ATL_IDS_SCSIZE          "Change the window size"
    ATL_IDS_SCMOVE          "Change the window position"
    ATL_IDS_SCMINIMIZE      "Reduce the window to an icon"
    ATL_IDS_SCMAXIMIZE      "Enlarge the window to full size"
    ATL_IDS_SCNEXTWINDOW    "Switch to the next document window"
    ATL_IDS_SCPREVWINDOW    "Switch to the previous document window"
    ATL_IDS_SCCLOSE         "Close the active window and prompts to save the documents"
END

STRINGTABLE
BEGIN
    ATL_IDS_SCRESTORE       "Restore the window to normal size"
    ATL_IDS_SCTASKLIST      "Activate Task List"
    ATL_IDS_MDICHILD        "Activate this window"
END

STRINGTABLE
BEGIN
    ATL_IDS_IDLEMESSAGE     "Ready"
END

STRINGTABLE
BEGIN
    ATL_IDS_MRU_FILE        "Open this document"
END

STRINGTABLE
BEGIN
    ID_CMD_RESET            "Reset\nReset"
    ID_CMD_STOP             "Stop\nStop"
    ID_FILE_DUMPDATA        "Dump data\nDump data"
    ID_CMD_SET_FREQ         "Set motor pwm\nSet motor pwm"
END

STRINGTABLE
BEGIN
    ID_CMD_SCAN             "Start Scan\nStart Scan"
END

#endif    // English (United States) resources
/////////////////////////////////////////////////////////////////////////////


