/** MEXPLUS: C++ Matlab MEX development kit.
 *
 * Kota Yamaguchi 2014
 */

#ifndef _mrpt_mexplus_H
#define _mrpt_mexplus_H

#include <mexplus/arguments.h>
#include <mexplus/dispatch.h>

// Redefine printf to avoid incompatibilities with the rest of MRPT libraries
#undef printf

#endif // _mrpt_mexplus_H
