/* +------------------------------------------------------------------------+
   |                     Mobile Robot Programming Toolkit (MRPT)            |
   |                          https://www.mrpt.org/                         |
   |                                                                        |
   | Copyright (c) 2005-2022, Individual contributors, see AUTHORS file     |
   | See: https://www.mrpt.org/Authors - All rights reserved.               |
   | Released under BSD License. See: https://www.mrpt.org/License          |
   +------------------------------------------------------------------------+ */
#pragma once

namespace mrpt::gui
{
/** Returns the default MRPT icon data, as GIMP header format. */
extern const char* default_mrpt_glfw_icon;
extern unsigned int default_mrpt_glfw_icon_width;
extern unsigned int default_mrpt_glfw_icon_height;
extern unsigned int default_mrpt_glfw_icon_transparent_color;

}  // namespace mrpt::gui
