/* +------------------------------------------------------------------------+
   |                     Mobile Robot Programming Toolkit (MRPT)            |
   |                          https://www.mrpt.org/                         |
   |                                                                        |
   | Copyright (c) 2005-2022, Individual contributors, see AUTHORS file     |
   | See: https://www.mrpt.org/Authors - All rights reserved.               |
   | Released under BSD License. See: https://www.mrpt.org/License          |
   +------------------------------------------------------------------------+ */
#pragma once
namespace mrpt::gui
{
class CDisplayWindow;
class CDisplayWindow3D;
class CDisplayWindowPlots;

class CWindowDialog;
class CWindowDialogPlots;
class C3DWindowDialog;
class CMyGLCanvas_DisplayWindow3D;

}  // namespace mrpt::gui
