/* +------------------------------------------------------------------------+
   |                     Mobile Robot Programming Toolkit (MRPT)            |
   |                          https://www.mrpt.org/                         |
   |                                                                        |
   | Copyright (c) 2005-2022, Individual contributors, see AUTHORS file     |
   | See: https://www.mrpt.org/Authors - All rights reserved.               |
   | Released under BSD License. See: https://www.mrpt.org/License          |
   +------------------------------------------------------------------------+ */
#pragma once

namespace mrpt::opengl
{
class CArrow;
class CAssimpModel;
class CAxis;
class CBox;
class CCamera;
class CColorBar;
class CCylinder;
class CDisk;
class CEllipsoid3D;
class CEllipsoidInverseDepth2D;
class CEllipsoidInverseDepth3D;
class CEllipsoidRangeBearing2D;
class CFrustum;
class CGridPlaneXY;
class CGridPlaneXZ;
class CMesh;
class CMesh3D;
class CMeshFast;
class COctoMapVoxels;
class COpenGLScene;
class COpenGLViewport;
class CPointCloud;
class CPointCloudColoured;
class CPolyhedron;
class CSetOfLines;
class CSetOfObjects;
class CSetOfTriangles;
class CSimpleLine;
class CSphere;
class CText;
class CText3D;
class CTexturedPlane;
class CVectorField2D;
class CVectorField3D;

struct TLightParameters;
}  // namespace mrpt::opengl
